/*
 * Convert a tokenized basic into listing.
 * When a BASIC program is stored on a Dragon DOS disk
 * then it is tokenized and there are 9 metadata bytes
 * at the start. This program converts a tokenized
 * BASIC program.
 */

#include <getopt.h>
#include <unistd.h>
#include "ddos.h"

// Reserved words - Start at 0x80
char *reserved[104] = {
 "FOR", "GO", "REM", "'", "ELSE", "IF", "DATA", "PRINT", "ON",
 "INPUT", "END", "NEXT", "DIM", "READ", "LET", "RUN", "RESTORE",
 "RETURN", "STOP", "POKE", "CONT", "LIST", "CLEAR", "NEW", "DEF",
 "CLOAD", "CSAVE", "OPEN", "CLOSE", "LLIST", "SET", "RESET", "CLS",
 "MOTOR", "SOUND", "AUDIO", "EXEC", "SKIPF", "DEL", "EDIT", "TRON",
 "TROFF", "LINE", "PCLS", "PSET", "PRESET", "SCREEN", "PCLEAR",
 "COLOR", "CIRCLE", "PAINT", "GET", "PUT", "DRAW", "PCOPY", "PMODE",
 "PLAY", "DLOAD", "RENUM", "TAB(", "TO", "SUB", "FN", "THEN", "NOT",
 "STEP", "OFF", "+", "-", "*", "/", "^", "AND", "OR", ">", "=", "<",
 "USING",
// Start from 0xCE
 "AUTO", "BACKUP", "BEEP", "BOOT", "CHAIN", "COPY", "CREATE", "DIR",
 "DRIVE", "DSKINIT", "FREAD", "FWRITE", "ERROR", "KILL", "LOAD", "MERGE",
 "PROTECT", "WAIT", "RENAME", "SAVE", "SREAD", "SWRITE", "VERIFY", "FROM",
 "FLREAD", "SWAP"
};

// Function tokens - all proceeded by 0xff to differentiate from operators
// Start at 0x80
char *funcs[41] = {
 "SGN", "INT", "ABS", "POS", "RND", "SQR", "LOG", "EXP", "SIN",
 "COS", "TAN", "ATN", "PEEK", "LEN", "STR$", "VAL", "ASC", "CHR$",
 "EOF", "JOYSTK", "FIX", "HEX$", "LEFT$", "RIGHT$", "MID$", "POINT",
 "INKEY$", "MEM", "VARPTR", "INSTR", "TIMER", "PPOINT", "STRING$", "USR",
// Start at 0xA2
 "LOF", "FREE", "ERL", "ERR", "HIMEM", "LOC", "FRE$"
};

void usage(char *prgname) {
    fprintf(stderr, "Usage: [-p] %s <tokenized basic file>\n", prgname);
    exit(1);
}

int main(int argc, char *argv[]) {

    int state, opt, prettify = 0;
    int spaceadd = 0;
    FILE *fp;
    struct fileinfo header;
    int c,d;

    if (argc < 2) {
        usage(argv[0]);
    }

    while ((opt = getopt(argc, argv, "p")) != -1) {
        switch(opt) {
        case 'p':
            prettify = 1;
            break;
        default:
            usage(argv[0]);
        }
    }

    if (optind >= argc) {
        fprintf(stderr, "Expected argument after options\n");
        exit(EXIT_FAILURE);
    }

    fp = fopen(argv[optind], "r");
    fread(&header, 9, 1, fp);
    if (header.startmark != 0x55 && header.endmark != 0xAA) {
        fprintf(stderr, "This is not a tokenized DRAGON BASIC file\n");
        exit(2);
    }
    state = 1;

    while (!feof(fp)) {
        switch (state) {
        case 1:  /* Pointer to next line */
            c = getc(fp);
            d = getc(fp);
            printf("\n");
            if (c == 0 && d == 0) state = 4;
            else state = 2;
            break;
        case 2:   /* Line number */
            c = getc(fp) & 0xFF;
            d = getc(fp) & 0xFF;
            printf("%d ", c * 256 + d);
            if (prettify) {
                spaceadd = 1;
            }
            state = 3;
            break;
        case 3:   /* Token or character */
            c = getc(fp) & 0xFF;
            if (c == 0) {
                state = 1;
                break;
            }
            if (c == 255) {
                c = getc(fp) & 0xFF;
                printf("%s", funcs[c - 128]);
            } else {
                if (c >= 128) {
                    if (prettify && !spaceadd) {
                        putchar(' ');
                    }
                    printf("%s", reserved[c - 128]);
                    if (prettify) {
                        putchar(' ');
                        spaceadd = 1;
                    }
                } else {
                    if (c != ' ' || !spaceadd)
                        printf("%c", c);
                    spaceadd = 0;
                }
            }
            break;
        default:  /* End of program */
            return 0;
        }
    }
}

